#include "headers.h"

int layers[4] = {LAYER1, LAYER2, LAYER3, LAYER4};

int layer[16] = {
	PA7, PA6, PA5, PA4,
	PA3, PA2, PA1, PA0,
	PC4, PC5, PC6, PC7,
	PD6, PC1, PC2, PC3
	};

int vx[4];

/* melodia curenta */
int current_id;
int *current_melody;
int *current_duration;
uint16_t current_size;
int16_t this_duration;
int model_layer;

/* iterez prin vectorul de note */
unsigned int k;

/* pentru modele */
int inner, outer;

/* initializari */
void IO_init(void) {

	/* pentru cub */
	DDRA |= 0xff;
	DDRC |= 0xff;
	DDRD |= (1 << PD7) | (1 << PD0) | (1 << PD1) | (1 << PD4) | (1 << PD6);

	/* PC0 ca intrare */
	DDRB &= ~(1 << PB2);
	PORTB |= (1 << PB2);
	
	/* initializere melodie */
	current_id = 0;
	current_melody = starwars;
	current_duration = starwars_duration;
	current_size = sizeof(starwars) / sizeof(int);
}

/*Functii pentru efectele cubului:
1. Toate layer-urile : functie folosita pentru a testa functionalitatea intregului cub.
2. Un singur layer. Am avut nevoie de functia asta la inceput pentru a verifica conectivitatea corecta la nivel hardware dintre cub
si tranzistori.
3. Aprinderea pe verticala a ledurilor.
4. Efectul de ploaie : functie ce aprinde ledurile incepand cu led-ul din layer-ul superior si coboara. Pentru a realiza asta, mai intai se aprind led-urile de pe layer-ul respectiv, se astepta, iar mai apo
se stinge ledul curent si se aprinde cel de pe urmatorul layer.
5.Efectul de miscare avion (traducere Google Translate) : se aprind led-urile fie pe layere si mai apoi se sting, fie pe verticale, iar mai apoi se sting.
6. Aprindere cub interior/exterior : aceste doua functii sunt folosite in efectul de miscare a led-urile pe ritmul muzicii. Nu mi-a reusit complet acest lucru.*/
/* functie folosita pentru a testa aprinderea intregului cub*/
void all_layers(void)
{
	PORTD |= (1<<LAYER1);	
	PORTD |= (1<<LAYER2);	
	PORTD |= (1<<LAYER3);	
	PORTD |= (1<<LAYER4);	
	
	PORTD |= (1 << PD6);
	PORTA = FULL_LAYER_A;
	PORTC = FULL_LAYER_C;
}

/* aprinde un layer intreg */
void one_layer(int i) {
	PORTD |= (1 << layers[i]);
	
	PORTD |= (1 << PD6);
	PORTA = FULL_LAYER_A;
	PORTC = FULL_LAYER_C;
}

/* aprinde o verticala */
void vertical_on(int i) {
	if (i >= 4)
		return;
	
	PORTD |= (1 << layers[0]) | (1 << layers[1]) | (1 << layers[2]) | (1 << layers[3]);
	
	PORTA |= (1 << layer[i]) | (1 << layer[i + 4]);
	if (i != 0) {
		PORTC |= (1 << layer[i + 8]) | (1 << layer[i + 12]);
	} else {
		PORTC |= (1 << layer[i + 8]);
		PORTD |= (1 << PD6);
	}
}

/* rain effect */
void rain(int iterations) {
	int i, k = 3, n, iter, x;
	
	for (iter = 0; iter < iterations; iter++) {
		/* random n */
		n = (int)((double)rand() / ((double)RAND_MAX + 1) * 4);
		for (i = 0; i < n; i++) {
			vx[i] = 0;
		}
		/* sunt pe ultimul nivel intai */
		PORTD |= (1 << layers[3]);
		for (i = 0; i < n; i++) {
			/* LED random */
			x = (int)((double)rand() / ((double)RAND_MAX + 1) * 16);
			vx[i] = x;
			if (x < 8) {
				PORTA |= (1 << layer[x]);
			} else {
				if (x == 12) {
					PORTD |= (1 << PD6);
				} else {
					PORTC |= (1 << layer[x]);
				}
			}
		}
		_delay_ms(80);
		k = 2;
		while (k >= 0) {
			/* trec la nivelul inferior */
			PORTD |= (1 << layers[k]);
			/* intai inchid toate LED-urile */
			PORTA = EMPTY_LAYER_A;
			PORTC = EMPTY_LAYER_C;
			PORTD &= ~(1 << PD6);
			/* aprind doar LED-urile aprinse pe nivelul cel mai de sus */
			for (i = 0; i < n; i++) {
				if (vx[i] != 0) {
					if (vx[i] < 8) {
						PORTA |= (1 << layer[vx[i]]);
					} else {
						if (vx[i] == 12) {
							PORTD |= (1 << PD6);
						} else {
							PORTC |= (1 << layer[vx[i]]);
						}
					}
				}
			}
			_delay_ms(80);
			k--;
		}
		/* dupa terminarea unei iteratii curat cubul */
		clear_all_layers();
		_delay_ms(80);
	}
}

/* planes moving */
void moving_planes(int dir) {
	int i;
	if (dir == 1) {
		/* de jos in sus */
		for (i = 3; i >= 0; i--) {
			/* aprind layerul respectiv */
			one_layer(i);
			_delay_ms(100);
			off_layer(i);
		}
		for (i = 0; i < 4; i++) {
			one_layer(i);
			_delay_ms(100);
			off_layer(i);
		}		
	} else {
		/* din fata in spate */
		for (i = 3; i >= 0; i--) {
			vertical_on(i);
			_delay_ms(100);
			vertical_off(i);
		}
		for (i = 0; i < 4; i++) {
			vertical_on(i);
			_delay_ms(100);
			vertical_off(i);
		}
	}
}

/* aprinde colturile */
void play_corner(void) {
	/* strat jos */
	PORTD |= (1 << LAYER1);
	/* strat sus */
	PORTD |= (1 << LAYER4);
	
	/* LED-urile */
	PORTC |= (1 << PC3);
	PORTD |= (1 << PD6);
	PORTA |= (1 << PA7) | ( 1 << PA4);
	_delay_ms(100);
}


/* aprinde cubul exterior */
void outer_cube(void) {
	vertical_on(0);
	vertical_on(3);
	PORTA |= (1 << PA6) | (1 << PA5);
	PORTC |= (1 << PC1) | (1 << PC2);
}
/* aprinde cubul interior */
void inner_cube(void) {
	PORTD |= (1 << PD1) | (1 << PD4);
	PORTA |= (1 << PA2) | (1 << PA1);
	PORTC |= (1 << PC5) | (1 << PC6);
}

/* functie folosita pentru a stinge intregul cub*/
void clear_all_layers(void)
{
	PORTD &= ~((1 << LAYER1));
	PORTD &= ~((1 << LAYER2));
	PORTD &= ~((1 << LAYER3));
	PORTD &= ~((1 << LAYER4));

	PORTD &= ~(1 << PD6);
	PORTA = EMPTY_LAYER_A;
	PORTC = EMPTY_LAYER_C;
}

/* stinge un layer */
void off_layer(int i) {
	PORTD &= ~(1 << layers[i]);

	PORTD &= ~(1 << PD6);
	PORTA = EMPTY_LAYER_A;
	PORTC = EMPTY_LAYER_C;
}

/* stinge o verticala */
void vertical_off(int i) {
	if (i >= 4)
		return;

	PORTD &= ~(1 << layers[0]); PORTD &= ~(1 << layers[1]); PORTD &= ~(1 << layers[2]); PORTD &= ~(1 << layers[3]);

	PORTA &= ~(1 << layer[i]); PORTA &= ~(1 << layer[i + 4]);
	if (i != 0) {
		PORTC &= ~(1 << layer[i + 8]); PORTC &= ~(1 << layer[i + 12]);
	}
	else {
		PORTC &= ~(1 << layer[i + 8]); PORTD &= ~(1 << PD6);
	}

}

/* stinge colturile */
void clear_corner(void) {
	PORTD &= ~(1 << LAYER1);
	PORTD &= ~(1 << LAYER4);

	PORTD &= ~(1 << PD6);
	PORTA = EMPTY_LAYER_A;
	PORTC = EMPTY_LAYER_C;
	_delay_ms(100);
}

/* stinge cubul exterior */
void clear_outer_cube(void) {
	vertical_off(0);
	vertical_off(3);
	PORTA &= ~((1 << PA6) | (1 << PA5));
	PORTC &= ~((1 << PC1) | (1 << PC2));
}

/* stinge cubul interior */
void clear_inner_cube(void) {
	PORTD &= ~((1 << PD1) | (1 << PD4));
	PORTA &= ~((1 << PA2) | (1 << PA1));
	PORTC &= ~((1 << PC5) | (1 << PC6));
}

/* initializeaza timer-ul 1 */
void timer1_init() {
    /* Fast PWM cu prag pe ICR1 si prescaler 1024 */
    TCCR1B |= (1 << CS10) | (1 << WGM13) | (1 << WGM12);
    TCCR1A |= (1 << WGM11);

	TCCR1A |= (1 << COM1B1);
	TCCR1A |= (1 << COM1A1);
	TCCR1A &= ~(1 << COM1B0);

	/* difuzor pe PD5*/
	DDRD |= (1 << PD5);

    ICR1 = 16000000 / current_melody[0];
    /* duty cycle 25% */
    OCR1B = ICR1 / 4;
    /* duty cycle-ul speaker (OC1A) 50% */
    OCR1A = ICR1 / 2;
}

/* aleg melodia. Am incercat sa fac sa fie mai multe melodii dar am avut un bug si nu am reusit sa fac butonul sa mearga la urmatoarea melodie, asa ca am ramas doar cu o singura melodie */
void choose_melody(void) {
	current_id = (current_id + 1) % 3;
	
	switch(current_id) {
		case STARWARS:
			current_melody = starwars;
			current_duration = starwars_duration;
			current_size = sizeof(starwars) / sizeof(int);
			break;
	}
}

/* in functie de amplitudine si de numarul modelului se aprind altfel LED-urile */
void model(int model_number) {	
	
	inner = FALSE;
	outer = FALSE;
	
	if(OCR1A < 9000) {
		if (model_number == 0) {
			one_layer(0);
		} else if (model_number == 1) {
			vertical_on(0);
		} else {
			inner_cube();
			inner = TRUE;
		}
		model_layer = LAYER1;
	} else if(OCR1A < 300) {	
		if (model_number == 0) {
			one_layer(1);
		} else if (model_number == 1) {
			vertical_on(1);
		} else {
			outer_cube();
			outer = TRUE;
		}
		model_layer = LAYER2;
	}
	else if(OCR1A < 15000) {
		if (model_number == 0) {
			one_layer(2);
		} else if (model_number == 1) {
			vertical_on(2);
		} else {
			inner_cube();
			inner = TRUE;
		}
		model_layer = LAYER3;
	}
	else if (OCR1A < 20000){
		if (model_number == 0) {
			one_layer(3);
		} else if (model_number == 1) {
			vertical_on(3);
		} else {
			outer_cube();
			outer = TRUE;
		}
		model_layer = LAYER4;
	} else {
		all_layers();
		model_layer = 8;
	}
}

/* sterge model */
void clear_model(int model_number) {
	if (model_layer == 8) {
		clear_all_layers();
	} else if (model_number == 0) {
		off_layer(model_layer);
	} else if (model_number == 1) {
		vertical_off(model_layer);
	} else {
		if (outer == TRUE) {
			clear_outer_cube();
		} else {
			clear_inner_cube();
		}
	}
}

int main (void)
{
	IO_init();	

	int iter_number = 0, init = 0, model_number = 0;
	char ok = FALSE;
	
	k = 1;
	
	while (1) {
		if (ok != TRUE) {
			/* intai animatiile */
			all_layers();
			if ((PINB & (1 << PB2)) == 0) {
				clear_all_layers();
				/* la apasarea butonului, aleg muzica */
				ok = TRUE;
				continue;
			}
			_delay_ms(500);
			clear_all_layers();

			/* al doilea efect */
			moving_planes(0);
			_delay_ms(500);
			clear_all_layers();
			if ((PINB & (1 << PB2)) == 0) {
				clear_all_layers();
				/* la apasarea butonului, aleg muzica */
				ok = TRUE;
				continue;
			}
			
			/*al treilea efect*/
			rain(10);
			_delay_ms(500);
			if ((PINB & (1 << PB2))  == 0) {
				clear_all_layers();
				/* la apasarea butonului, aleg muzica */
				ok = TRUE;
				continue;
			}
			/*al patrulea efect */
			moving_planes(1);
			if ((PINB & (1 << PB2)) == 0) {
				clear_all_layers();
				/* la apasarea butonului, aleg muzica */
				ok = TRUE;
				continue;
			}
			
			play_corner();
			clear_corner();
			
			if ((PINB & (1 << PB2)) == 0) {
				clear_all_layers();
				/* la apasarea butonului, aleg muzica */
				ok = TRUE;
				continue;
			}
		} else {
			if (init == 0) {
				/* pornesc redarea audio */
				timer1_init();

				init = 1;
			}
			
			/* in functie de ICR1 canta notele */
			ICR1 = F_CPU / current_melody[k];
			OCR1A = ICR1 / 2;
			OCR1B = ICR1 / 4;
			
			/* alegem modelul */
			model(model_number);

			_delay_ms(100);
			k++;
			if (k == current_size) {
				k = 0;
			}
			
			/* sting ce am aprins inainte */
			clear_model(model_number);
			
			/* la apasarea pe buton se schimba melodiile */
			if ((PINC & (1 << PB2)) == 0) {
				choose_melody();
			}
		}
	}

	return 0;
}